# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
#
# Flask-Resources is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Schema that's able to handle loading data from MultiDicts."""

import marshmallow as ma
from werkzeug.datastructures import MultiDict


class MultiDictSchema(ma.Schema):
    """MultiDict aware schema used for loading e.g. request.args."""

    class Meta:
        """Throw away unknown values."""

        unknown = ma.EXCLUDE

    LIST_TYPES = [
        ma.fields.List,
    ]

    @classmethod
    def is_list_field(cls, field):
        """Method used to determine if a field is a list type field.

        This is used to unpack the MultiDict into a normal dictionary.
        """
        return any((isinstance(field, t) for t in cls.LIST_TYPES))

    @ma.pre_load
    def flatten_multidict(self, data, **kwargs):
        """Flatten a MultiDict into a normal dictionary."""
        if not isinstance(data, MultiDict):
            return data
        data = data.to_dict(flat=False)

        for name, field in self.fields.items():
            if name in data and not self.is_list_field(field):
                data[name] = data[name][0]
        return data



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# Flask-Resources is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Serializers required interfaces."""


class SerializerMixin:
    """Serializer Interface."""

    def serialize_object(self, obj, response_ctx=None, *args, **kwargs):
        """Serialize a single object according to the response ctx.

        The object type must implement ``SerializableMixin``.
        """
        raise NotImplementedError()

    def serialize_object_list(self, obj_list, response_ctx=None, *args, **kwargs):
        """Serialize a list of objects according to the response ctx.

        Each object type of the list should implement ``SerializableMixin``.
        """
        raise NotImplementedError()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
# Copyright (C) 2020 Northwestern University.
#
# Flask-Resources is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Library for easily implementing REST APIs."""

from .headers import HeadersParser, headers_parser
from .schema import MultiDictSchema
from .url_args import URLArgsParser, url_args_parser

__all__ = (
    "headers_parser",
    "HeadersParser",
    "MultiDictSchema",
    "url_args_parser",
    "URLArgsParser",
)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# Flask-Resources is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Views."""

from .collections import ItemView, ListView
from .singleton import SingletonView

__all__ = ("ItemView", "ListView", "SingletonView")



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
# Copyright (C) 2020 Northwestern University.
#
# Flask-Resources is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Flask Resources module to create REST APIs."""

from flask import request
from werkzeug.exceptions import HTTPException

from ..context import resource_requestctx
from .base import BaseView


class ListView(BaseView):
    """List view representation.

    Allows searching and creating an item in the list.
    """

    def __init__(self, *args, **kwargs):
        """Constructor."""
        super(ListView, self).__init__(*args, **kwargs)
        # is defined by response_handler decorator
        self.response_handler = None

    @property
    def resource_method(self):
        """Returns string of resource method according to request.method."""
        if request.method == "GET":
            return "search"
        if request.method == "POST":
            return "create"
        return ""

    def get(self, *args, **kwargs):
        """Search the collection."""
        # TODO: Make it so that you don't have to return a tuple. See issue #55
        return self.response_handler.make_list_response(
            *self.resource.search(*args, **kwargs)
        )

    def post(self, *args, **kwargs):
        """Create an item in the collection."""
        return self.response_handler.make_item_response(
            *self.resource.create(*args, **kwargs)  # data is passed in the context
        )

    def put(self, *args, **kwargs):
        """Update the collection."""
        return self.response_handler.make_list_response(
            *self.resource.update_all(*args, **kwargs)
        )

    def delete(self, *args, **kwargs):
        """Delete the collection."""
        return self.response_handler.make_list_response(
            *self.resource.delete_all(*args, **kwargs)
        )


class ItemView(BaseView):
    """Item view representation.

    Allows reading, (partial) updating and deleting an item.
    """

    def __init__(self, *args, **kwargs):
        """Constructor."""
        super(ItemView, self).__init__(*args, **kwargs)
        # is defined by response_handler decorator
        self.response_handler = None

    @property
    def resource_method(self):
        """Returns string of resource method according to request.method."""
        if request.method == "GET":
            return "read"
        if request.method == "PUT":
            return "update"
        if request.method == "PATCH":
            return "partial_update"
        if request.method == "DELETE":
            return "delete"
        return ""

    def get(self, *args, **kwargs):
        """Get."""
        return self.response_handler.make_item_response(
            *self.resource.read(*args, **kwargs)
        )

    def put(self, *args, **kwargs):
        """Put."""
        return self.response_handler.make_item_response(
            *self.resource.update(*args, **kwargs)  # data is passed in the context
        )

    def patch(self, *args, **kwargs):
        """Patch."""
        return self.response_handler.make_item_response(
            *self.resource.partial_update(*args, **kwargs)
        )

    def delete(self, *args, **kwargs):
        """Delete."""
        return self.response_handler.make_item_response(
            *self.resource.delete(*args, **kwargs)
        )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2021 CERN.
# Copyright (C) 2020-2021 Northwestern University.
#
# Flask-Resources is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Header parser."""

from flask import request

from .base import RequestParser, request_parser_decorator


class HeadersParser(RequestParser):
    """Headers parser."""

    def load_data(self):
        """Load data from headers."""
        return {k.lower().replace("-", "_"): v for (k, v) in request.headers.items()}


headers_parser = request_parser_decorator(
    HeadersParser,
    # Attribute name on the resource config
    "request_headers_parser",
    # Attribute name on the resource_requestctx
    "headers",
)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# Flask--Resources is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Pytest configuration.

See https://pytest-invenio.readthedocs.io/ for documentation on which test
fixtures are available.
"""

import pytest
from flask import Flask

from flask_resources.context import resource_requestctx
from flask_resources.resources import CollectionResource, Resource, ResourceConfig


class CustomResourceConfig(ResourceConfig):
    """Custom resource configuration."""

    item_route = "/custom/<id>"
    list_route = "/custom/"


class CustomResource(CollectionResource):
    """Custom resource implementation."""

    default_config = CustomResourceConfig

    def __init__(self, *args, **kwargs):
        """Constructor."""
        super(CustomResource, self).__init__(*args, **kwargs)
        self.db = {}

    def search(self):
        """Search."""
        query = resource_requestctx.url_args.get("q", "")
        resp = []
        for key, value in self.db.items():
            if query in key or query in value:
                resp.append({"id": key, "content": value})

        return resp, 200

    def create(self):
        """Create."""
        obj = resource_requestctx.request_content
        self.db[obj["id"]] = obj["content"]
        return self.db, 201

    def read(self):
        """Read."""
        _id = resource_requestctx.route["id"]
        return {"id": _id, "content": self.db[_id]}, 200

    def delete(self):
        """Delete."""
        _id = resource_requestctx.route["id"]
        if _id in self.db:
            del self.db[_id]
        return {}, 200

    def update_all(self):
        """Delete."""
        for obj in resource_requestctx.request_content:
            self.db[obj["id"]] = obj["content"]
        return resource_requestctx.request_content, 200


@pytest.fixture(scope="module")
def app():
    """Application factory fixture."""
    app_ = Flask(__name__)

    default_bp = Resource().as_blueprint("default_resource")
    app_.register_blueprint(default_bp)
    custom_bp = CustomResource().as_blueprint("custom_resource")
    app_.register_blueprint(custom_bp)

    return app_



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# Flask-Resources is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""JSON Deserializer."""

import json

from .deserializers import DeserializerMixin


class JSONDeserializer(DeserializerMixin):
    """JSON Deserializer."""

    def deserialize_data(self, data, *args, **kwargs):
        """Deserializes the input data into json."""
        return json.loads(data) if data else {}



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# Flask--Resources is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Errors test module."""

import pytest
from flask import Flask
from werkzeug.exceptions import HTTPException

from flask_resources.errors import (
    HTTPJSONException,
    InvalidContentType,
    MIMETypeNotAccepted,
    create_errormap_handler,
)
from flask_resources.resources import CollectionResource, Resource, ResourceConfig


# These classes are in the file because they are under test too
class CustomInvalidContentType(HTTPJSONException):

    code = 415
    description = "My custom exception."


class CustomMIMETypeNotAccepted(HTTPJSONException):

    code = 406

    def __init__(self, description, **kwargs):
        super(CustomMIMETypeNotAccepted, self).__init__(**kwargs)
        self.description = "CustomMIMETypeNotAccepted: {}".format(description)


class CustomErrorMapConfig(ResourceConfig):
    """Custom resource configuration."""

    item_route = "/custom/<id>"
    list_route = "/custom/"
    error_map = {
        InvalidContentType: create_errormap_handler(CustomInvalidContentType()),
        MIMETypeNotAccepted: create_errormap_handler(
            lambda e: CustomMIMETypeNotAccepted(e.description)
        ),
    }


class CustomConfig(ResourceConfig):
    """Custom resource configuration."""

    item_route = "/custom/<id>"
    list_route = "/custom/"


class CustomResource(CollectionResource):
    """Custom resource implementation."""

    default_config = CustomConfig


@pytest.fixture(scope="module")
def app():
    """Application factory fixture."""
    app_ = Flask(__name__)
    custom_bp = CustomResource().as_blueprint("custom_resource")
    app_.register_blueprint(custom_bp)
    return app_


@pytest.fixture(scope="function")
def app_with_error_mapping():
    """Application factory fixture."""
    app_ = Flask(__name__)

    CustomErrorMapConfig.error_map.update()
    custom_bp = CustomResource(config=CustomErrorMapConfig).as_blueprint(
        "custom_resource"
    )
    app_.register_blueprint(custom_bp)
    return app_


def test_404_if_no_route(client):
    response = client.get("/foo")
    assert response.status_code == 404

    response = client.post("/foo")
    assert response.status_code == 404

    response = client.put("/foo")
    assert response.status_code == 404

    response = client.patch("/foo")
    assert response.status_code == 404

    response = client.delete("/foo")
    assert response.status_code == 404


def test_405_if_route_but_no_method(client):
    # TODO: Only implement MethodView methods for defined Resource methods. Only then
    #       can we test that all methods return 405 (and not just the ones this library
    #       didn't define in its MethodViews
    # NOTE: For now we just test the methods that this library
    #       didn't define in its MethodViews
    # Item-level
    response = client.post("/custom/1")
    assert response.status_code == 405


def test_406_if_route_method_but_unsupported_accept(client):
    # NOTE: By default we don't accept any mimetype at all
    headers = {"accept": "application/marcxml+garbage"}

    # List-level
    response = client.get("/custom/", headers=headers)
    assert response.status_code == 406

    response = client.post("/custom/", headers=headers)
    assert response.status_code == 406

    # Item-level
    response = client.get("/custom/1", headers=headers)
    assert response.status_code == 406

    response = client.put("/custom/1", headers=headers)
    assert response.status_code == 406

    response = client.patch("/custom/1", headers=headers)
    assert response.status_code == 406

    response = client.delete("/custom/1", headers=headers)
    assert response.status_code == 406


def test_406_with_mapped_exceptions(app_with_error_mapping):
    client = app_with_error_mapping.test_client()

    # NOTE: By default we don't accept any mimetype at all
    headers = {"accept": "application/marcxml+garbage"}
    expected_message = "CustomMIMETypeNotAccepted: Invalid 'Accept' header"

    # List-level
    response = client.get("/custom/", headers=headers)
    assert response.status_code == 406
    assert expected_message in response.json["message"]

    response = client.post("/custom/", headers=headers)
    assert response.status_code == 406
    assert expected_message in response.json["message"]

    # Item-level
    response = client.get("/custom/1", headers=headers)
    assert response.status_code == 406
    assert expected_message in response.json["message"]

    response = client.put("/custom/1", headers=headers)
    assert response.status_code == 406
    assert expected_message in response.json["message"]

    response = client.patch("/custom/1", headers=headers)
    assert response.status_code == 406
    assert expected_message in response.json["message"]

    response = client.delete("/custom/1", headers=headers)
    assert response.status_code == 406
    assert expected_message in response.json["message"]


def test_415_if_route_method_accept_but_unsupported_content_type(client):
    # NOTE: Right now, we rely entirely on headers for payload assessment.
    headers = {"accept": "application/json", "content-type": "application/json+garbage"}

    # NOTE: We only test methods that accept a payload
    # List-level
    response = client.post("/custom/", headers=headers)
    assert response.status_code == 415

    # Item-level
    response = client.put("/custom/1", headers=headers)
    assert response.status_code == 415

    response = client.patch("/custom/1", headers=headers)
    assert response.status_code == 415


def test_415_with_mapped_exception(app_with_error_mapping):
    client = app_with_error_mapping.test_client()

    # NOTE: Right now, we rely entirely on headers for payload assessment.
    headers = {"accept": "application/json", "content-type": "application/json+garbage"}

    # NOTE: We only test methods that accept a payload
    # List-level
    response = client.post("/custom/", headers=headers)
    assert response.status_code == 415
    assert response.json["message"] == "My custom exception."

    # Item-level
    response = client.put("/custom/1", headers=headers)
    assert response.status_code == 415
    assert response.json["message"] == "My custom exception."

    response = client.patch("/custom/1", headers=headers)
    assert response.status_code == 415
    assert response.json["message"] == "My custom exception."



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
# Copyright (C) 2020 Northwestern University.
#
# Flask-Resources is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Loader module."""

from functools import wraps

from flask import request

from .context import resource_requestctx
from .errors import InvalidContentType


def select_deserializer(resource_method, loader_or_loaders):
    """Returns deserializer corresponding to situation."""
    if isinstance(loader_or_loaders, RequestLoader):
        return loader_or_loaders, loader_or_loaders.deserializer

    loader = loader_or_loaders.get(resource_method)
    if loader:
        return loader, loader.deserializer
    else:
        raise InvalidContentType(allowed_mimetypes=loader_or_loaders.keys())


def request_loader(f):
    """Decorator that loads/deserializes the request data."""

    @wraps(f)
    def wrapper(self, *args, **kwargs):
        """Wrapping method.

        :params self: Item/List/SingletonView instance
        """
        # Checking Content-Type is the responsibility of the loader
        loaders = self.resource.config.request_loaders

        request_content_type = request.content_type

        if not request_content_type:
            request_content_type = self.resource.config.default_content_type

        if request_content_type not in loaders:
            raise InvalidContentType(allowed_mimetypes=loaders.keys())

        loader, deserializer = select_deserializer(
            self.resource_method, loaders[request_content_type]
        )

        resource_requestctx.request_content = deserializer.deserialize_data(
            loader.load_data()
        )

        return f(self, *args, **kwargs)

    return wrapper


class LoaderMixin:
    """Loader interface."""

    def load_data(self):
        """Return request data."""
        raise NotImplementedError()

    def load_item_request(self, *args, **kwargs):
        """Load a request concerning an existing item."""
        raise NotImplementedError()

    def load_search_request(self, *args, **kwargs):
        """Load a search request."""
        raise NotImplementedError()


class RequestLoader(LoaderMixin):
    """Loaded request representation.

    Loads the content from the request.
    """

    def __init__(self, deserializer=None, args_parser=None, *args, **kwargs):
        """Constructor."""
        self.deserializer = deserializer
        self.args_parser = args_parser

    def load_data(self):
        """Return request data."""
        return request.data

    def load_item_request(self, *args, **kwargs):
        """Build response headers."""
        return {"request_content": self.deserializer.deserialize_data(self.data)}



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# Flask-Resources is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Library for easily implementing REST APIs."""

from .resources import CollectionResource, Resource, ResourceConfig, SingletonResource
from .version import __version__

__all__ = (
    "__version__",
    "Resource",
    "ResourceConfig",
    "CollectionResource",
    "SingletonResource",
)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
# Copyright (C) 2020 Northwestern University.
#
# Flask-Resources is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Resource Request context."""

from functools import wraps

from flask import g
from werkzeug.local import LocalProxy


#
# Proxy to the current resource context
#
def _get_context():
    """Get the resource request context from the g object."""
    if hasattr(g, "resource_requestctx"):
        return g.resource_requestctx
    raise RuntimeError("Working outside of resource request context.")


resource_requestctx = LocalProxy(_get_context)
"""Proxy to the resource's request context"""


#
# Resource context
#
class ResourceRequestCtx(object):
    """Context manager for the resource context.

    The resource request context encodes information about the currently
    executing request for a given resource, such as:

    - The mimetype selected by the content negotiation.
    - The content type of the request payload
    """

    def __init__(self, accept_mimetype=None, request_args=None, data=None):
        """Initialize the resource context."""
        self.accept_mimetype = accept_mimetype
        self.request_args = request_args
        self.request_content = data
        self.route = {}

    def __enter__(self):
        """Push the resource context manager on the current request."""
        g.resource_requestctx = self

    def __exit__(self, type, value, traceback):
        """Pop the resource context manager from the current request."""
        del g.resource_requestctx

    def update(self, values):
        """Update the context fields present in the received dictionary `values`."""
        for field, value in values.items():
            self.__setattr__(field, value)


def with_resource_requestctx(f):
    """Wrap in resource request context."""

    @wraps(f)
    def inner(self, *args, **kwargs):
        with ResourceRequestCtx():
            return f(self, *args, **kwargs)

    return inner


def with_route(f):
    """Wrapper to extract route arguments into resource context.

    NOTE: This removes the need for *args, **kwargs passing in Resource methods
    TODO: This should be temporary because the whole decorator approach should be
          revisited.
    """

    @wraps(f)
    def inner(*args, **kwargs):
        resource_requestctx.route = kwargs
        return f(*args)

    return inner



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
# Copyright (C) 2020 Northwestern University.
#
# Flask-Resources is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Test deserialization."""

import marshmallow as ma
from flask import Flask, request
from speaklater import make_lazy_string

from flask_resources.serializers import JSONSerializer, MarshmallowJSONSerializer


def _(s):
    return make_lazy_string(lambda: s)


def test_lazy_strings_are_serialized():
    serializer = JSONSerializer()
    lazy = {"key": _("Lazy")}
    list_lazy = [{"key": _("Lazy1")}, {"key": _("Lazy2")}]

    assert '{"key": "Lazy"}' == serializer.serialize_object(lazy)
    assert '[{"key": "Lazy1"}, {"key": "Lazy2"}]' == serializer.serialize_object_list(
        list_lazy
    )


def test_prettyprint():
    app = Flask("test")
    with app.test_request_context("/?prettyprint=1"):
        serializer = JSONSerializer()
        assert '{\n  "key": "1"\n}' == serializer.serialize_object({"key": "1"})


def test_marhsmallow_serializer():
    class TestSchema(ma.Schema):
        title = ma.fields.Str(data_key="test")

    s = MarshmallowJSONSerializer(item_schema=TestSchema, list_schema=TestSchema)
    assert s.serialize_object({"title": "a"}) == '{"test": "a"}'
    assert s.serialize_object_list({"title": "a"}) == '{"test": "a"}'



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
# Copyright (C) 2020 Northwestern University.
#
# Flask-Resources is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Library for easily implementing REST APIs."""

import warnings
from functools import wraps

import marshmallow as ma

from ..context import resource_requestctx


def select_request_parser(
    resource_method, config_parser_or_parsers, default_parser_cls
):
    """Returns request parser corresponding to situation."""
    if isinstance(config_parser_or_parsers, default_parser_cls):
        return config_parser_or_parsers

    return config_parser_or_parsers.get(
        resource_method,
        default_parser_cls(allow_unknown=False),  # default "parse nothing" parser
    )


def request_parser_decorator(parser_cls, config_attr, context_attr):
    """Decorator that parses the request object for the view."""

    def decorator(f):
        @wraps(f)
        def inner(self, *args, **kwargs):
            parser = select_request_parser(
                self.resource_method,
                getattr(self.resource.config, config_attr),
                parser_cls,
            )
            setattr(resource_requestctx, context_attr, parser.parse())
            return f(self, *args, **kwargs)

        return inner

    return decorator


class _ExcludeSchema(ma.Schema):
    class Meta:
        unknown = ma.EXCLUDE


class RequestParser:
    """Request parser.

    Base class for parsing request.
    """

    DEFAULT_SCHEMA_CLASS = _ExcludeSchema
    """The schema class to use if a dictionary is passed."""

    def __init__(self, schema=None, allow_unknown=None):
        """Constructor.

        :param schema: A marshmallow schema class or a mapping from keys to
        fields.
        """
        self._schema = schema
        self.allow_unknown = allow_unknown
        if allow_unknown is not None:
            warnings.warn(
                "The allow_unknown keyword argument is deprecated and has no "
                "effect. Use unknown values control on marshmallow instead.",
                DeprecationWarning,
            )

    @property
    def schema(self):
        """Build the schema class."""
        if self._schema is None:
            return None
        if isinstance(self._schema, dict):
            cls_ = self.DEFAULT_SCHEMA_CLASS
            if self.allow_unknown is True:
                # allow_unknown is deprecated (see __init__ above). Here we
                # add backward compatibility, so that when you pass a dict
                # you can use the allow_unknown keyword. Instead of using
                # allow_unknown you should now simply create a schema instead
                # and add the Meta class as show below:
                class _IncludeSchema(cls_):
                    class Meta:
                        unknown = ma.INCLUDE

                cls_ = _IncludeSchema

            return cls_.from_dict(self._schema)()
        else:
            return self._schema()

    def load_data(self):
        """Load data from request.

        This should be overwritten in subclasses to implement loading of data
        from the querystring, form values, headers etc.
        """
        raise NotImplementedError

    def parse(self):
        """Parse the request data."""
        if self.schema is None:
            return {}
        return self.schema.load(self.load_data())



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# Flask-Resources is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Deserializers."""

from .deserializers import DeserializerMixin
from .json import JSONDeserializer

__all__ = ("JSONDeserializer", "DeserializerMixin")



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
# Copyright (C) 2020 Northwestern University.
#
# Flask-Resources is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Resource view."""

from flask import Blueprint
from marshmallow import ValidationError

from .deserializers import JSONDeserializer
from .errors import create_errormap_handler
from .loaders import RequestLoader
from .parsers import HeadersParser, URLArgsParser
from .responses import Response
from .serializers import JSONSerializer
from .views import ItemView, ListView, SingletonView

ITEM_VIEW_SUFFIX = "_item"
LIST_VIEW_SUFFIX = "_list"


class ResourceConfig:
    """Base resource configuration."""

    request_loaders = {
        "application/json": RequestLoader(deserializer=JSONDeserializer())
    }
    response_handlers = {"application/json": Response(JSONSerializer())}
    item_route = "/resources/<id>"
    list_route = "/resources/"
    request_url_args_parser = URLArgsParser()
    request_headers_parser = HeadersParser()
    default_content_type = "application/json"
    default_accept_mimetype = "application/json"
    error_map = {}


class Resource:
    """Resource controller interface."""

    default_config = ResourceConfig

    def __init__(self, config=None):
        """Initialize the base resource."""
        # TODO: The config should be checked to see that it is consistent. See issue #57
        self.config = config or self.default_config
        self.bp_name = None

    # Primary interface
    def search(self):
        """Perform a search over the items."""
        return [], 200

    def create(self):
        """Create an item."""
        return {}, 200

    def read(self):
        """Read an item."""
        return {}, 200

    def update(self):
        """Update an item."""
        return {}, 200

    def partial_update(self):
        """Partial update an item."""
        return {}, 200

    def delete(self):
        """Delete an item."""
        return {}, 200

    # Secondary interface
    def as_blueprint(self, name, **bp_kwargs):
        """Create blueprint and register rules only for the RecordResource."""
        self.bp_name = name
        blueprint = Blueprint(name, __name__, **bp_kwargs)

        for rule in self.create_url_rules(name):
            blueprint.add_url_rule(**rule)

        for exc_or_code, error_handler in self.create_errormap_handlers():
            blueprint.register_error_handler(exc_or_code, error_handler)

        return blueprint

    def create_url_rules(self, bp_name):
        """Create url rules."""
        return [
            {
                "rule": self.config.item_route,
                "view_func": ItemView.as_view(
                    name=f"{bp_name}",
                    resource=self,
                ),
            }
        ]

    def create_errormap_handlers(self):
        """Create error handlers."""
        return self.config.error_map.items()


class CollectionResource(Resource):
    """CollectionResource."""

    def update_all(self):
        """Delete an item."""
        return [], 200

    def delete_all(self):
        """Delete an item."""
        return [], 200

    def create_url_rules(self, bp_name):
        """Create url rules."""
        return [
            {
                "rule": self.config.item_route,
                "view_func": ItemView.as_view(
                    name=f"{bp_name}{ITEM_VIEW_SUFFIX}",
                    resource=self,
                ),
            },
            {
                "rule": self.config.list_route,
                "view_func": ListView.as_view(
                    name=f"{bp_name}{LIST_VIEW_SUFFIX}",
                    resource=self,
                ),
            },
        ]


class SingletonResource(Resource):
    """SingletonResource."""

    def create_url_rules(self, bp_name):
        """Create url rules."""
        return [
            {
                "rule": self.config.list_route,
                "view_func": SingletonView.as_view(
                    name=f"{bp_name}",
                    resource=self,
                ),
            }
        ]



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
# Copyright (C) 2020 Northwestern University.
#
# Flask-Resources is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Parses test module."""

import pytest
from flask import Flask
from marshmallow import ValidationError, fields
from werkzeug.exceptions import HTTPException

from flask_resources.context import resource_requestctx
from flask_resources.errors import (
    HTTPJSONException,
    create_errormap_handler,
    handle_http_exception,
)
from flask_resources.parsers import HeadersParser, URLArgsParser
from flask_resources.resources import CollectionResource, ResourceConfig

# NOTE: mimetype headers have to be provided in general, but the args parsing is not
#       dependent on them specifically directly i.e. it could have been xml if xml
#       were implemented
HEADERS = {"content-type": "application/json", "accept": "application/json"}


# These classes are in the file because they are under test too
class UniversalParserConfig(ResourceConfig):
    """Resource configuration with URLArgsParser for all endpoints."""

    item_route = "/universal/<id>"
    list_route = "/universal"
    # Because an URLArgsParser is passed directly, it is applied to all endpoints
    request_url_args_parser = URLArgsParser(
        {"num": fields.Int(), "lang": fields.String(missing="")}
    )
    request_headers_parser = HeadersParser({"content_type": fields.String()})


class MethodParserConfig(ResourceConfig):
    """Resource configuration with URLArgsParser for a specific endpoint."""

    item_route = "/method/<id>"
    list_route = "/method"
    # Because an URLArgsParser is passed to "search", it is only applied
    # to that endpoint
    request_url_args_parser = {
        "search": URLArgsParser(
            {"num": fields.Int(), "lang": fields.String(missing="")}, allow_unknown=True
        )
    }

    request_headers_parser = {
        "update": HeadersParser({"content_type": fields.String()})
    }

    error_map = {
        ValidationError: create_errormap_handler(
            HTTPJSONException(
                code=400,
                description="Validation error.",
            )
        )
    }


class ParserResource(CollectionResource):
    """Resource implementation to test parsers."""

    default_config = UniversalParserConfig

    def search(self, *args, **kwargs):
        """Perform a search over the items."""
        return {**resource_requestctx.url_args, **resource_requestctx.headers}, 200

    def create(self, *args, **kwargs):
        """Create an item."""
        return {**resource_requestctx.url_args, **resource_requestctx.headers}, 200

    def read(self, *args, **kwargs):
        """Read an item."""
        return {**resource_requestctx.url_args, **resource_requestctx.headers}, 200

    def update(self, *args, **kwargs):
        """Update an item."""
        return {**resource_requestctx.url_args, **resource_requestctx.headers}, 200

    def partial_update(self, *args, **kwargs):
        """Partial update an item."""
        return {**resource_requestctx.url_args, **resource_requestctx.headers}, 200

    def delete(self, *args, **kwargs):
        """Delete an item."""
        return {**resource_requestctx.url_args, **resource_requestctx.headers}, 200


@pytest.fixture(scope="module")
def app():
    """Application factory fixture."""
    app_ = Flask(__name__)
    custom_bp = ParserResource().as_blueprint("universal_resource")
    app_.register_blueprint(custom_bp)
    custom_bp = ParserResource(MethodParserConfig).as_blueprint("method_resource")
    app_.register_blueprint(custom_bp)
    app_.register_error_handler(HTTPException, handle_http_exception)
    return app_


def test_universal_parser_parses_one_endpoint(client):
    response = client.get("/universal/1?num=10", headers=HEADERS)

    assert response.status_code == 200
    assert response.json["num"] == 10
    assert response.json["lang"] == ""
    assert response.json["content_type"] == "application/json"


def test_universal_parser_parses_all_endpoints(client):
    # search
    response = client.get("/universal?num=10", headers=HEADERS)
    assert response.status_code == 200
    assert response.json["num"] == 10
    assert response.json["lang"] == ""
    assert response.json["content_type"] == "application/json"

    # create
    response = client.post("/universal?num=10", headers=HEADERS)
    assert response.status_code == 200
    assert response.json["num"] == 10
    assert response.json["lang"] == ""
    assert response.json["content_type"] == "application/json"

    # read -- covered above

    # update
    response = client.put("/universal/1?num=10", headers=HEADERS)
    assert response.status_code == 200
    assert response.json["num"] == 10
    assert response.json["lang"] == ""
    assert response.json["content_type"] == "application/json"

    # partial update
    response = client.patch("/universal/1?num=10", headers=HEADERS)
    assert response.status_code == 200
    assert response.json["num"] == 10
    assert response.json["lang"] == ""
    assert response.json["content_type"] == "application/json"

    # delete
    response = client.delete("/universal/1?num=10", headers=HEADERS)
    assert response.status_code == 200
    assert response.json["num"] == 10
    assert response.json["lang"] == ""
    assert response.json["content_type"] == "application/json"


def test_method_parser_parses_only_method_endpoint(client):
    # search
    response = client.get("/method?num=10", headers=HEADERS)
    assert response.status_code == 200
    assert response.json["num"] == 10
    assert response.json["lang"] == ""

    # Other methods should return an empty json
    # Passing query strings when no parser for them is totally fine
    # they are just not extracted.

    # create
    response = client.post("/method?num=10", headers=HEADERS)
    assert response.status_code == 200
    assert response.json == {}

    # read
    response = client.get("/method/1?num=10", headers=HEADERS)
    assert response.status_code == 200
    assert response.json == {}

    # update
    response = client.put("/method/1?num=10", headers=HEADERS)
    assert response.status_code == 200
    assert response.json["content_type"] == "application/json"

    # partial update
    response = client.patch("/method/1?num=10", headers=HEADERS)
    assert response.status_code == 200
    assert response.json == {}

    # delete
    response = client.delete("/method/1?num=10", headers=HEADERS)
    assert response.status_code == 200
    assert response.json == {}


def test_parser_raises_400(client):
    response = client.get("/method?num=NotANumber", headers=HEADERS)

    assert response.status_code == 400
    assert response.json["status"] == 400
    assert response.json["message"] is not None


def test_parser_includes_or_excludes_unknown_args(client):
    response = client.get("/method?num=10&foo=20&bar=1&foo=10", headers=HEADERS)
    assert response.status_code == 200
    assert response.json == {"num": 10, "lang": "", "foo": ["20", "10"], "bar": ["1"]}

    response = client.get("/universal?num=10&foo=20&bar=1&foo=10", headers=HEADERS)
    assert response.status_code == 200
    assert response.json == {"content_type": "application/json", "num": 10, "lang": ""}



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
# Copyright (C) 2020 Northwestern University.
#
# Flask-Resources is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

from werkzeug.datastructures import MIMEAccept
from werkzeug.http import parse_accept_header

from flask_resources.content_negotiation import ContentNegotiator


# Test content negotiation by Accept header
# NOTE: By scoping down we remove the need to check for HTTP method
def test_choose_provided_and_accepted_mimetype():
    # Should choose mimetype that is accepted by client and served by server
    server_mimetypes = ["application/json", "application/marcxml+xml"]
    client_mimetypes = parse_accept_header(
        "text/plain,application/json,*/*", MIMEAccept
    )

    assert "application/json" == ContentNegotiator.match_by_accept(
        server_mimetypes, client_mimetypes
    )

    client_mimetypes = parse_accept_header(
        "text/plain,application/marcxml+xml,*/*", MIMEAccept
    )

    assert "application/marcxml+xml" == ContentNegotiator.match_by_accept(
        server_mimetypes, client_mimetypes
    )


def test_favour_specificity_over_quality():
    # favour more specific but lower quality mimetype over
    # less specific (e.g. wildcard) but higher quality
    server_mimetypes = ["application/json", "application/marcxml+xml"]
    client_mimetypes = parse_accept_header(
        "text/plain, application/json;q=0.5, */*", MIMEAccept
    )

    assert "application/json" == ContentNegotiator.match_by_accept(
        server_mimetypes, client_mimetypes
    )


def test_favour_quality_over_same_specificity():
    server_mimetypes = ["application/json", "application/marcxml+xml"]
    client_mimetypes = parse_accept_header(
        "application/json;q=0.5, application/marcxml+xml", MIMEAccept
    )

    assert "application/marcxml+xml" == ContentNegotiator.match_by_accept(
        server_mimetypes, client_mimetypes
    )

    client_mimetypes = parse_accept_header(
        "application/marcxml+xml;q=0.4, application/json;q=0.6", MIMEAccept
    )

    assert "application/json" == ContentNegotiator.match_by_accept(
        server_mimetypes, client_mimetypes
    )


def test_choose_default_if_no_match_and_wildcard_accepted():
    # choose default if no match and client accepts wildcard
    server_mimetypes = ["application/json", "application/marcxml+xml"]
    client_mimetypes = parse_accept_header("text/plain,*/*", MIMEAccept)

    assert "application/json" == ContentNegotiator.match_by_accept(
        server_mimetypes, client_mimetypes, default="application/json"
    )


def test_choose_none_if_no_match_and_wildcard_not_accepted():
    server_mimetypes = ["application/json", "application/marcxml+xml"]
    client_mimetypes = parse_accept_header("text/plain", MIMEAccept)

    mime_type = ContentNegotiator.match_by_accept(
        server_mimetypes, client_mimetypes, default="application/json"
    )

    assert mime_type is None


def test_choose_default_if_nothing_accepted():
    server_mimetypes = ["application/json", "application/marcxml+xml"]
    client_mimetypes = parse_accept_header("", MIMEAccept)

    assert "application/json" == ContentNegotiator.match_by_accept(
        server_mimetypes, client_mimetypes, default="application/json"
    )


# Test content negotiation by URL argument
# NOTE: By scoping down we remove the need to check for HTTP method
def test_choose_query_mimetype():
    formats_map = {
        "json": "application/json",
        "marcxml": "application/marcxml+xml",
    }
    fmt = "marcxml"  # this is the query

    assert "application/marcxml+xml" == ContentNegotiator.match_by_format(
        formats_map, fmt
    )

    fmt = "json"

    assert "application/json" == ContentNegotiator.match_by_format(formats_map, fmt)

    fmt = "foo"

    mime_type = ContentNegotiator.match_by_format(formats_map, fmt)

    assert mime_type is None


# Test top-level ContentNegotiator.match
def test_favour_query_mimetype_over_header_mimetype():
    server_mimetypes = ["application/json", "application/marcxml+xml"]
    client_mimetypes = parse_accept_header("application/json", MIMEAccept)
    formats_map = {
        "json": "application/json",
        "marcxml": "application/marcxml+xml",
    }
    fmt = "marcxml"

    assert "application/marcxml+xml" == ContentNegotiator.match(
        server_mimetypes, client_mimetypes, formats_map, fmt
    )

    client_mimetypes = parse_accept_header("application/marcxml+xml", MIMEAccept)
    fmt = "json"

    assert "application/json" == ContentNegotiator.match(
        server_mimetypes, client_mimetypes, formats_map, fmt
    )


def test_favour_header_mimetype_if_no_query_mimetype():
    server_mimetypes = ["application/json", "application/marcxml+xml"]
    client_mimetypes = parse_accept_header("application/json", MIMEAccept)
    formats_map = {
        "json": "application/json",
        "marcxml": "application/marcxml+xml",
    }
    fmt = None

    assert "application/json" == ContentNegotiator.match(
        server_mimetypes, client_mimetypes, formats_map, fmt
    )

    formats_map = {}
    fmt = "marcxml"

    assert "application/json" == ContentNegotiator.match(
        server_mimetypes, client_mimetypes, formats_map, fmt
    )


def test_choose_default_if_no_query_and_no_header():
    server_mimetypes = ["application/json", "application/marcxml+xml"]
    client_mimetypes = parse_accept_header("", MIMEAccept)
    formats_map = {
        "json": "application/json",
        "marcxml": "application/marcxml+xml",
    }
    fmt = None

    assert "application/json" == ContentNegotiator.match(
        server_mimetypes, client_mimetypes, formats_map, fmt, default="application/json"
    )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
# Copyright (C) 2020 Northwestern University.
#
# Flask-Resources is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Custom JSON encoder."""


from flask.json import JSONEncoder
from speaklater import is_lazy_string


class CustomJSONEncoder(JSONEncoder):
    """JSONEncoder for our custom needs.

    - Knows to force translate lazy translation strings
    """

    def default(self, obj):
        """Override parent's default."""
        if is_lazy_string(obj):
            return str(obj)
        return super().default(obj)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
# Copyright (C) 2020 Northwestern University.
#
# Flask-Resources is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""JSON serializer."""

import json

from flask import request

from ..json_encoder import CustomJSONEncoder
from .serializers import SerializerMixin


def flask_request_options():
    """Options to pretty print the JSON."""
    if request and request.args.get("prettyprint"):
        return {
            "indent": 2,
            "sort_keys": True,
        }
    return {}


class JSONSerializer(SerializerMixin):
    """JSON serializer implementation."""

    def __init__(self, encoder=None, options=None):
        """Initialize the JSONSerializer."""
        self._options = options or flask_request_options
        self._encoder = encoder or CustomJSONEncoder

    @property
    def dumps_options(self):
        """Support adding options for the dumps() method."""
        return self._options() if callable(self._options) else self._options

    @property
    def encoder(self):
        """Support overriding the JSONEncoder used for serialization."""
        # We let classes through as-is
        if isinstance(self._encoder, type):
            return self._encoder
        elif callable(self._encoder):
            return self._encoder()
        return self._encoder

    def serialize_object(self, obj):
        """Dump the object into a json string."""
        encoder = self.encoder
        return json.dumps(obj, cls=encoder, **self.dumps_options)

    def serialize_object_list(self, obj_list):
        """Dump the object list into a json string."""
        return json.dumps(obj_list, cls=self.encoder, **self.dumps_options)


class MarshmallowJSONSerializer(JSONSerializer):
    """JSON serializing using Marshmallow to transform output."""

    def __init__(self, item_schema=None, list_schema=None, **kwargs):
        """Initialize the serializer."""
        self._item_schema_cls = item_schema
        self._list_schema_cls = list_schema
        super().__init__(**kwargs)

    def serialize_object(self, obj):
        """Dump the object into a JSON string."""
        return super().serialize_object(self.dump_item(obj))

    def serialize_object_list(self, obj_list):
        """Dump the object list into a JSON string."""
        return super().serialize_object_list(self.dump_list(obj_list))

    def dump_item(self, obj):
        """Dump the object with extra information."""
        return self._item_schema_cls().dump(obj)

    def dump_list(self, obj_list):
        """Dump the list of objects with extra information."""
        ctx = {
            "item_schema_cls": self._item_schema_cls,
        }
        return self._list_schema_cls(context=ctx).dump(obj_list)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# Flask-Resources is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Version information for Flask-Resources.

This file is imported by ``flask_resources.__init__``,
and parsed by ``setup.py``.
"""

__version__ = "0.5.0"



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
# Copyright (C) 2020 Northwestern University.
#
# Flask-Resources is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Resources test module."""

import json


def test_base_resource(client):
    """Test the default resource."""
    # TODO: Only implement MethodView methods for defined Resource methods. Only then
    #       can we test that all methods return 405 (and not just the ones this library
    #       didn't define in its MethodViews
    headers = {"content-type": "application/json", "accept": "application/json"}

    # GET/Read a resource
    response = client.get("/resources/1234-ABCD", headers=headers)
    assert response.status_code == 200
    assert response.json == {}

    # PUT/Edit a resource
    obj_json = json.dumps({"id": "1234-ABCD", "content": "something new"})
    response = client.put("/resources/1234-ABCD", data=obj_json, headers=headers)
    assert response.status_code == 200
    assert response.json == {}

    # PATCH/Partial edit a resource
    ops_json = json.dumps(
        {"op": "replace", "path": "/content", "value": "something patched"}
    )
    response = client.patch("/resources/1234-ABCD", data=obj_json, headers=headers)
    assert response.status_code == 200
    assert response.json == {}

    # DELETE/remove a resource
    response = client.delete("/resources/1234-ABCD", headers=headers)
    assert response.status_code == 200
    assert response.json == {}


def test_custom_resource(client):
    """Test a custom resource."""
    headers = {"content-type": "application/json", "accept": "application/json"}

    # The resource returns a list, empty for now
    resource_obj = client.get("/custom/", headers=headers)

    assert resource_obj.status_code == 200
    assert len(resource_obj.json) == 0

    # Create a resource object
    obj_content = {"id": "1234-ABCD", "content": "test resource obj content"}
    obj_json = json.dumps(obj_content)
    resource_obj = client.post("/custom/", data=obj_json, headers=headers)
    assert resource_obj.status_code == 201

    # Search for the previously created obj
    resource_obj = client.get("/custom/", headers=headers)
    assert resource_obj.status_code == 200
    assert len(resource_obj.json) == 1
    assert resource_obj.json[0]["id"] == "1234-ABCD"

    # Get the previously created obj
    resource_obj = client.get("/custom/1234-ABCD", headers=headers)
    assert resource_obj.status_code == 200
    assert resource_obj.json["id"] == "1234-ABCD"

    # Get the previously created obj using default headers
    resource_obj = client.get("/custom/1234-ABCD")
    assert resource_obj.status_code == 200
    assert resource_obj.json["id"] == "1234-ABCD"

    # Delete the previously created obj
    resource_obj = client.delete("/custom/1234-ABCD", headers=headers)
    assert resource_obj.status_code == 200
    assert resource_obj.json == {}

    # PUT/Edit a list of resource units
    obj_json = json.dumps([{"id": "1234-ABCD", "content": "updated content"}])
    response = client.put("/custom/", data=obj_json, headers=headers)
    assert response.status_code == 200
    assert response.json == [{"id": "1234-ABCD", "content": "updated content"}]

    # DELETE/remove a list of resource units
    response = client.delete("/custom/", headers=headers)
    assert response.status_code == 200
    assert response.json == []



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# Flask-Resources is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Response module."""

from functools import wraps

from flask import make_response

from .context import resource_requestctx


def response_handler(f):
    """Decorator that sets the response_handler on the view."""

    @wraps(f)
    def wrapper(self, *args, **kwargs):
        """Wrapping method.

        :params self: Item/List/SingletonView instance
        """
        accept_mimetype = resource_requestctx.accept_mimetype
        # NOTE: accept_mimetype is in handlers for sure at this point
        handlers = self.resource.config.response_handlers
        self.response_handler = handlers[accept_mimetype]
        return f(self, *args, **kwargs)

    return wrapper


class ResponseMixin:
    """Response interface."""

    def make_headers(self, content=None):
        """Build response headers."""
        return {"content-type": resource_requestctx.accept_mimetype}

    def make_item_response(self, content, code):
        """Builds a response."""
        raise NotImplementedError()

    def make_list_response(self, content, code):
        """Builds a response."""
        raise NotImplementedError()

    def make_error_response(self, error):
        """Builds an error response."""
        raise NotImplementedError()


class Response(ResponseMixin):
    """Response implementation.

    Builds up a reponse for a single or list of objects.
    """

    def __init__(self, serializer=None):
        """Constructor."""
        self.serializer = serializer

    def make_item_response(self, content, code=200):
        """Builds a response for a single object."""
        # https://flask.palletsprojects.com/en/1.1.x/api/#flask.Flask.make_response
        # (body, status, header)

        return make_response(
            "" if content is None else self.serializer.serialize_object(content),
            code,
            self.make_headers(content=content),
        )

    def make_list_response(self, content, code=200):
        """Builds a response for a list of objects."""
        # https://flask.palletsprojects.com/en/1.1.x/api/#flask.Flask.make_response
        # (body, status, header)

        return make_response(
            "" if content is None else self.serializer.serialize_object_list(content),
            code,
            self.make_headers(),
        )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# Flask-Resources is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Exceptions used in Flask Resources module."""

import json

from flask import g
from werkzeug.exceptions import HTTPException

from .json_encoder import CustomJSONEncoder


def create_errormap_handler(map_func_or_exception):
    """Creates a resource errormap handler.

    The handler is used to map business logic exceptions to
    REST exceptions. The original exceptions is being stored
    in the `__original_exc__` attribute of the mapped exception.

    :param map_func_or_exception: Function or exception to map orginally
    raised exception to a `flask_resources.errors.HTTPJSONException`.
    """

    def error_handler(e):
        if isinstance(map_func_or_exception, HTTPJSONException):
            mapped_exc = map_func_or_exception
        else:
            mapped_exc = map_func_or_exception(e)
        mapped_exc.__original_exc__ = e
        return mapped_exc.get_response()

    return error_handler


handle_http_exception = create_errormap_handler(
    lambda exc: HTTPJSONException(code=exc.code, description=exc.description)
)


class HTTPJSONException(HTTPException):
    """HTTP Exception delivering JSON error responses."""

    errors = None

    def __init__(self, code=None, errors=None, **kwargs):
        """Initialize HTTPJSONException."""
        super(HTTPJSONException, self).__init__(**kwargs)
        if errors is not None:
            self.errors = errors
        if code is not None:
            self.code = code

    @property
    def name(self):
        """The status name."""
        return type(self).__name__

    def get_errors(self):
        """Get errors.

        :returns: A list containing the errors.
        """
        return [e for e in self.errors] if self.errors else None

    def get_description(self, environ=None):
        """Returns an unescaped description."""
        return self.description

    def get_headers(self, environ=None):
        """Get a list of headers."""
        return [("Content-Type", "application/json")]

    def get_body(self, environ=None):
        """Get the request body."""
        body = {"status": self.code, "message": self.get_description(environ)}

        errors = self.get_errors()
        if errors:
            body["errors"] = errors

        # TODO: Revisit how to integrate error monitoring services. See issue #56
        # Temporarily kept for expediency and backward-compatibility
        if self.code and (self.code >= 500) and hasattr(g, "sentry_event_id"):
            body["error_id"] = str(g.sentry_event_id)

        return json.dumps(body, cls=CustomJSONEncoder)


class MIMETypeException(HTTPJSONException):
    """Error for when an invalid Content-Type is provided."""

    header_name = None

    def __init__(self, allowed_mimetypes=None, **kwargs):
        """Initialize exception."""
        super(MIMETypeException, self).__init__(**kwargs)
        self.allowed_mimetypes = allowed_mimetypes
        self.description = "Invalid '{0}' header. Expected one of: {1}".format(
            self.header_name, ", ".join(allowed_mimetypes)
        )


class MIMETypeNotAccepted(MIMETypeException):
    """Error for when an invalid `Accept` header is provided."""

    code = 406
    header_name = "Accept"


class InvalidContentType(MIMETypeException):
    """Error for when an invalid `Content-Type` header is provided."""

    code = 415
    header_name = "Content-Type"



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# Flask-Resources is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Library for easily implementing REST APIs."""


class FlaskResources(object):
    """Flask-Resources extension."""

    def __init__(self, app=None):
        """Constructor."""
        if app:
            self.init_app(app)

    def init_app(self, app):
        """Flask application initialization."""
        app.extensions["flask-resources"] = self



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# Flask-Resources is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Serializers."""

from .json import JSONSerializer, MarshmallowJSONSerializer
from .serializers import SerializerMixin

__all__ = ("JSONSerializer", "MarshmallowJSONSerializer", "SerializerMixin")



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
# Copyright (C) 2020 Northwestern University.
#
# Flask-Resources is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Parses test module."""

import pytest
from flask import Flask
from marshmallow import fields

from flask_resources.context import resource_requestctx
from flask_resources.parsers import HeadersParser
from flask_resources.resources import CollectionResource, ResourceConfig

# NOTE: mimetype headers have to be provided in general, but the args parsing is not
#       dependent on them specifically directly i.e. it could have been xml if xml
#       were implemented
HEADERS = {"content-type": "application/json", "accept": "application/json"}


# These classes are in the file because they are under test too
class HeaderParserConfig(ResourceConfig):
    """Resource configuration for headers test (nothing special for now)."""

    item_route = "/headers/<id>"
    list_route = "/headers"
    request_headers_parser = HeadersParser(
        {
            "content_type": fields.String(),
            "accept": fields.String(),
        }
    )


class HeadersResource(CollectionResource):
    """Resource implementation to test headers."""

    default_config = HeaderParserConfig

    def _headers_to_dict(self):
        return dict(resource_requestctx.headers.items())

    def search(self, *args, **kwargs):
        """Perform a search over the items."""
        return self._headers_to_dict(), 200

    def create(self, *args, **kwargs):
        """Create an item."""
        return self._headers_to_dict(), 200

    def read(self, *args, **kwargs):
        """Read an item."""
        # import pdb; pdb.set_trace()
        return self._headers_to_dict(), 200

    def update(self, *args, **kwargs):
        """Update an item."""
        return self._headers_to_dict(), 200

    def partial_update(self, *args, **kwargs):
        """Partial update an item."""
        return self._headers_to_dict(), 200

    def delete(self, *args, **kwargs):
        """Delete an item."""
        return self._headers_to_dict(), 200


@pytest.fixture(scope="module")
def app():
    """Application factory fixture."""
    app_ = Flask(__name__)
    custom_bp = HeadersResource().as_blueprint("headers_resource")
    app_.register_blueprint(custom_bp)
    return app_


@pytest.fixture(scope="module")
def expected_headers():
    return {
        "accept": "application/json",
        "content_type": "application/json",
    }


def test_headers_read_endpoint(client, expected_headers):
    response = client.get("/headers/1", headers=HEADERS)
    assert response.status_code == 200
    assert expected_headers == response.json


def test_headers_search_endpoint(client, expected_headers):
    response = client.get("/headers", headers=HEADERS)
    assert response.status_code == 200
    assert expected_headers == response.json


def test_headers_create_endpoint(client, expected_headers):
    response = client.post("/headers", headers=HEADERS)
    assert response.status_code == 200
    assert expected_headers == response.json


def test_headers_update_endpoint(client, expected_headers):
    response = client.put("/headers/1", headers=HEADERS)
    assert response.status_code == 200
    assert expected_headers == response.json


def test_headers_partial_update_endpoint(client, expected_headers):
    response = client.patch("/headers/1", headers=HEADERS)
    assert response.status_code == 200
    assert expected_headers == response.json


def test_headers_delete_endpoint(client, expected_headers):
    response = client.delete("/headers/1", headers=HEADERS)
    assert response.status_code == 200
    assert expected_headers == response.json



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
# Copyright (C) 2020 Northwestern University.
#
# Flask-Resources is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Deserializers required interfaces."""


class DeserializerMixin:
    """Deserializer Interface."""

    def deserialize_data(self, data, *args, **kwargs):
        """Deserializes an object."""
        raise NotImplementedError()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
# Copyright (C) 2020 Northwestern University.
#
# Flask-Resources is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Test deserialization."""

import pytest
from flask import Flask

from flask_resources.context import resource_requestctx
from flask_resources.deserializers import JSONDeserializer
from flask_resources.loaders import RequestLoader
from flask_resources.resources import CollectionResource, ResourceConfig


# These classes are in the file because they are under test too
class UniversalDeserializerConfig(ResourceConfig):
    """Resource configuration with deserializer for all endpoints."""

    item_route = "/universal/<id>"
    list_route = "/universal"
    # Because a deserializer is passed directly, it is applied to all endpoints
    request_loaders = {
        # TODO: Rename config and Remove RequestLoader
        "application/json": RequestLoader(deserializer=JSONDeserializer())
    }


class MethodDeserializerConfig(ResourceConfig):
    """Resource configuration with deserializer for a specific endpoint."""

    item_route = "/method/<id>"
    list_route = "/method"
    # Because a deserializer is passed to "search", it is only applied to that endpoint
    request_loaders = {
        # TODO: Rename config and Remove RequestLoader
        "application/json": {"search": RequestLoader(deserializer=JSONDeserializer())}
    }


class DeserializerResource(CollectionResource):
    """Resource implementation to test deserialization."""

    default_config = UniversalDeserializerConfig

    def search(self, *args, **kwargs):
        """Return sent payload."""
        return resource_requestctx.request_content, 200

    def delete(self, *args, **kwargs):
        """Return sent payload."""
        return resource_requestctx.request_content, 200


@pytest.fixture(scope="module")
def app():
    """Application factory fixture."""
    app_ = Flask(__name__)
    custom_bp = DeserializerResource().as_blueprint("universal_resource")
    app_.register_blueprint(custom_bp)
    custom_bp = DeserializerResource(MethodDeserializerConfig).as_blueprint(
        "method_resource"
    )
    app_.register_blueprint(custom_bp)

    return app_


def test_search_endpoint_deserializes_data(client):
    headers = {"content-type": "application/json", "accept": "application/json"}

    response = client.get("/universal", json={"q": "foo=bar"}, headers=headers)

    assert response.status_code == 200
    assert response.json == {"q": "foo=bar"}

    # Works well even if no data
    response = client.get("/universal", headers=headers)

    assert response.status_code == 200
    assert response.json == {}


def test_delete_endpoint_deserializes_data(client):
    headers = {"content-type": "application/json", "accept": "application/json"}

    response = client.delete("/universal/1", json={"tombstone": "baz"}, headers=headers)

    assert response.status_code == 200
    assert response.json == {"tombstone": "baz"}

    # Works well even if no data
    response = client.delete("/universal/1", headers=headers)

    assert response.status_code == 200
    assert response.json == {}


def test_method_deserializer_deserializes_search_only(client):
    headers = {"content-type": "application/json", "accept": "application/json"}

    response = client.get("/method", json={"q": "foo=bar"}, headers=headers)

    assert response.status_code == 200
    assert response.json == {"q": "foo=bar"}

    response = client.delete("/method/1", json={"tombstone": "baz"}, headers=headers)

    assert response.status_code == 415
    assert response.json["status"] == 415
    assert response.json["message"] is not None



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# Flask-Resources is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Flask Resources module to create REST APIs."""

from flask import request

from .base import BaseView


class SingletonView(BaseView):
    """Parse request args and create a request context.

    Note that the resource route should contain the `<id>` param.
    """

    def __init__(self, *args, **kwargs):
        """Constructor."""
        super(SingletonView, self).__init__(*args, **kwargs)
        # is defined by response_handler decorator
        self.response_handler = None

    @property
    def resource_method(self):
        """Returns string of resource method according to request.method."""
        if request.method == "GET":
            return "read"
        if request.method == "PUT":
            return "update"
        if request.method == "PATCH":
            return "partial_update"
        if request.method == "DELETE":
            return "delete"
        if request.method == "POST":
            return "create"
        return ""

    def post(self, *args, **kwargs):
        """Post."""
        return self.response_handler.make_item_response(
            *self.resource.create(*args, **kwargs)  # data is passed in the context
        )

    def get(self, *args, **kwargs):
        """Get."""
        return self.response_handler.make_item_response(
            *self.resource.read(*args, **kwargs)  # data is passed in the context
        )

    def put(self, *args, **kwargs):
        """Put."""
        return self.response_handler.make_item_response(
            *self.resource.update(*args, **kwargs)  # data is passed in the context
        )

    def patch(self, *args, **kwargs):
        """Patch."""
        return self.response_handler.make_item_response(
            *self.resource.partial_update(
                *args, **kwargs
            )  # data is passed in the context
        )

    def delete(self, *args, **kwargs):
        """Delete."""
        return self.response_handler.make_item_response(
            *self.resource.delete(*args, **kwargs)  # data is passed in the context
        )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# Flask-Resources is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Flask Resources module to create REST APIs."""

import os

from setuptools import find_packages, setup

readme = open("README.rst").read()
history = open("CHANGES.rst").read()

tests_require = [
    "black>=20.8b1,<20.9b0",
    "check-manifest>=0.42",
    "coverage>=5.3,<6",
    "pytest>=6,<7",
    "pytest-cov>=2.10.1",
    "pytest-flask>=1.0.0",
    "pytest-isort>=1.2.0",
    "pytest-mock>=1.6.0",
    "pytest-pycodestyle>=2.2.0",
    "pytest-pydocstyle>=2.2.0",
]

extras_require = {
    "docs": ["Sphinx>=3",],
    "tests": tests_require,
}

extras_require["all"] = []
for reqs in extras_require.values():
    extras_require["all"].extend(reqs)

setup_requires = ["Babel>=1.3"]

install_requires = [
    "Flask~=1.1.2",
    "marshmallow~=3.0",
    "speaklater>=1.3,<2.0",
]

packages = find_packages()


# Get the version string. Cannot be done with import!
g = {}
with open(os.path.join("flask_resources", "version.py"), "rt") as fp:
    exec(fp.read(), g)
    version = g["__version__"]

setup(
    name="flask-resources",
    version=version,
    description=__doc__,
    long_description=readme + "\n\n" + history,
    keywords="flask TODO",
    license="MIT",
    author="CERN",
    author_email="info@inveniosoftware.org",
    url="https://github.com/inveniosoftware/flask-resources",
    packages=packages,
    zip_safe=False,
    include_package_data=True,
    platforms="any",
    entry_points={},
    extras_require=extras_require,
    install_requires=install_requires,
    setup_requires=setup_requires,
    tests_require=tests_require,
    classifiers=[
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Development Status :: 1 - Planning",
    ],
)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
# Copyright (C) 2020 Northwestern University.
#
# Flask-Resources is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Flask Resources module to create REST APIs.

The views responsabilities are:
- Parse request arguments.
- Build a request context for the resource.
"""

from flask.views import MethodView

from ..content_negotiation import content_negotiation
from ..context import with_resource_requestctx, with_route
from ..loaders import request_loader
from ..parsers import headers_parser, url_args_parser
from ..responses import response_handler


class BaseView(MethodView):
    """Base view."""

    resource_decorators = [
        request_loader,
        url_args_parser,
        with_route,
        response_handler,
        content_negotiation,
        headers_parser,
        with_resource_requestctx,
    ]
    """Resource-specific decorators to be applied to the views."""

    def __init__(self, resource, *args, **kwargs):
        """Constructor."""
        super(BaseView, self).__init__(*args, **kwargs)
        self.resource = resource

    def dispatch_request(self, *args, **kwargs):
        """Dispatch request after applying resource decorators."""
        view = MethodView.dispatch_request

        for decorator in self.resource_decorators:
            view = decorator(view)

        return view(self, *args, **kwargs)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
# Copyright (C) 2020 Northwestern University.
#
# Flask-Resources is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Library for easily implementing REST APIs."""

from flask import request

from .base import RequestParser, request_parser_decorator
from .schema import MultiDictSchema


class URLArgsParser(RequestParser):
    """URL query string parser."""

    DEFAULT_SCHEMA_CLASS = MultiDictSchema

    def __init__(self, *args, **kwargs):
        """Initialize the arguments parser.

        :param schema: The marshmallow schema class or a dictionary of fields.
        :param as_dict: If ``False``, the schema is passed a MultiDict object and
            must be able to handle it (see
            ``flask_resources.parsers.MultiDictSchema``). If True, the schema
            is passed a dictionary where the values are lists.
        """
        self.as_dict = kwargs.pop("as_dict", False)
        super().__init__(*args, **kwargs)

    def load_data(self):
        """Load data from the request args."""
        if self.as_dict:
            return request.args.to_dict(flat=False)
        else:
            return request.args


url_args_parser = request_parser_decorator(
    URLArgsParser,
    # Attribute name on the resource config
    "request_url_args_parser",
    # Attribute name on the resource_requestctx
    "url_args",
)
